@@ -2,14 +2,15 @@
 # All rights reserved
 # Licensed under a 3-clause BSD style license (see LICENSE)
 import re
-from collections import defaultdict
+import weakref
 
 import pywintypes
 import win32pdh
 
 from ....errors import ConfigTypeError, ConfigValueError
+from .constants import PDH_NO_DATA
 from .transform import NATIVE_TRANSFORMERS, TRANSFORMERS
-from .utils import construct_counter_path, format_instance, get_counter_value
+from .utils import construct_counter_path, get_counter_value, get_counter_values, validate_path
 
 
 class PerfObject:
@@ -35,6 +36,8 @@ class PerfObject:
         if not isinstance(self.tag_name, str):
             raise ConfigTypeError(f'Option `tag_name` for performance object `{self.name}` must be a string')
 
+        # List of regex patterns to filter multi-instance counters AFTER ALL the data
+        # is collected and retrieved from PDH layer
         include_patterns = config.get('include', [])
         if not isinstance(include_patterns, list):
             raise ConfigTypeError(f'Option `include` for performance object `{self.name}` must be an array')
@@ -49,6 +52,8 @@ class PerfObject:
 
             self.include_pattern = re.compile('|'.join(include_patterns))
 
+        # List of regex patterns to filter multi-instance counters AFTER ALL data
+        # is collected and retrieved from PDH layer
         exclude_patterns = config.get('exclude', [])
         if not isinstance(exclude_patterns, list):
             raise ConfigTypeError(f'Option `exclude` for performance object `{self.name}` must be an array')
@@ -63,6 +68,22 @@ class PerfObject:
             final_exclude_patterns.extend(exclude_patterns)
             self.exclude_pattern = re.compile('|'.join(final_exclude_patterns))
 
+        # List of wildcards or instance name directly to filter multi-instance counters by PDH layer itself.
+        # Thus it is faster and and less resource intensive than regex-based include filtering.
+        include_wildcards = config.get('include_fast', [])
+        if not isinstance(include_wildcards, list):
+            raise ConfigTypeError(f'Option `include_fast` for performance object `{self.name}` must be an array')
+        elif not include_wildcards:
+            self.include_wildcards = None
+        else:
+            for i, pattern in enumerate(include_wildcards, 1):
+                if not isinstance(pattern, str):
+                    raise ConfigTypeError(
+                        f'Pattern #{i} of option `include_fast` for performance object `{self.name}` must be a string'
+                    )
+
+            self.include_wildcards = include_wildcards
+
         instance_counts = config.get('instance_counts', {})
         if not isinstance(instance_counts, dict):
             raise ConfigTypeError(f'Option `instance_counts` for performance object `{self.name}` must be a mapping')
@@ -91,6 +112,54 @@ class PerfObject:
         elif not self.counters_config:
             raise ConfigValueError(f'Option `counters` for performance object `{self.name}` is required')
 
+        # Temporary and not documented, at least at this point, configuration value which controls which API
+        # will be invoked to get multiple instance values. If the configuration is `true`
+        # (duplicate_instances_exist: true), which is default, then we will invoke Windows
+        # PdhGetFormattedCounterArrayW() API and process its output in pure Python. If not, we will invoke
+        # win32pdh.GetFormattedCounterArray() which is 5-10x times faster than using
+        # PdhGetFormattedCounterArrayW(), but it will lose duplicate/non-unique instances. There are a few
+        # reasons for this configuration and its default value (true). Please note that PDH API calls
+        # contribute only ~1% of overall performance overhead and making them 5-10x faster will not make
+        # overall that much faster and most likely the gain will be lost in the non-deterministic noise. Still
+        # these kind improvements are not wasteful:
+        #
+        #     * This new multi-instance wildcard-based implementation is 2-10+x faster than preceding
+        #       per-instance based implementation (advantage is bigger the more instances and counters
+        #       are involved). Thus the slower API calls are effectively on par regarding overall
+        #       performance impact. It is true, its full performance advantage of wildcard-based
+        #       implementation will not be realized until we can switch to an enhanced version of
+        #       win32pdh.GetFormattedCounterArray(). Again, in overall performance overhead the gain
+        #       will be small peanuts.
+        #
+        #     * We plan to file a ticket and perhaps assist to enhance win32pdh.GetFormattedCounterArray()
+        #       function to handle non-unique instances. When it will be implemented we can remove slower
+        #       Python implementation and this configuration variable along with it.
+        #
+        #     * In general, these calls are relatively fast. Typically they take microseconds or lower
+        #       milliseconds (unless there are more than a few thousand instances). Multiplying by 10 will
+        #       not make a drastic impact on overall performance since it may be call only every 15 seconds.
+        #
+        #     * Microsoft requests counter provider developers to avoid using duplicates. It is required
+        #       for Performance Counters Provider V2 but it is not guaranteed in general for V1 counters.
+        #       Microsoft's own counters seem to not allow duplicates with well known exceptions for
+        #       "Process" performance objects. Even though it is a waste at this point in the virtual
+        #       majority of the performance counters, since most of them will not have duplicate instance
+        #       names to use the slower call which handles duplicates, still we would like to avoid changing
+        #       default behavior and lose duplicates.
+        #
+        #     * This configuration at this point is provided just in case to have a fallback route if
+        #       performance impact of the slower call could be problematic in some circumstances.
+        #       To activate faster call, which cannot handle duplicate, use duplicate_instances_exist: false
+        #
+        #     * Because we are planning to pull out this configuration after relevant changes in win32pdh
+        #       we are not planning at this time to make it public in the sense of adding it to configuration
+        #       sample file, or explicitly documenting it elsewhere except perhaps internal Kb articles.
+        self.duplicate_instances_exist = config.get('duplicate_instances_exist', True)
+        if not isinstance(self.duplicate_instances_exist, bool):
+            raise ConfigTypeError(
+                f'Option `duplicate_instances_exist` for performance object `{self.name}` must be an true or false'
+            )
+
         # We'll configure on the first run because it's necessary to figure out whether the
         # object contains single or multi-instance counters, see:
         # https://docs.microsoft.com/en-us/windows/win32/perfctrs/object-and-counter-design
@@ -101,52 +170,56 @@ class PerfObject:
         for counter in self.counters:
             counter.collect()
 
-    def refresh(self):
-        # https://docs.microsoft.com/en-us/windows/win32/api/pdh/nf-pdh-pdhenumobjectitemsa
-        # https://mhammond.github.io/pywin32/win32pdh__EnumObjectItems_meth.html
-        counters, instances = win32pdh.EnumObjectItems(
-            None, self.connection.server, self.name, win32pdh.PERF_DETAIL_WIZARD
-        )
-        if not self.counters:
-            self._configure_counters(counters, instances)
+        # Collect "instance" metrics directly from a MultiCounter counter object
+        if self.has_multiple_instances and len(self.counters) > 0:
+            counter = next(iter(self.counters))
+            total_instance_count, monitored_instance_count, unique_instance_count = counter.get_instance_count()
 
-        counters = set(counters)
-        total_instance_count = 0
-        monitored_instance_count = 0
-        unique_instance_counts = defaultdict(int)
-        for instance in instances:
-            total_instance_count += 1
-            if self._instance_excluded(instance):
-                continue
-
-            monitored_instance_count += 1
-            unique_instance_counts[instance] += 1
-
-        if self.has_multiple_instances:
             if self.instance_count_total_metric is not None:
                 self.check.gauge(self.instance_count_total_metric, total_instance_count, tags=self.tags)
             if self.instance_count_monitored_metric is not None:
                 self.check.gauge(self.instance_count_monitored_metric, monitored_instance_count, tags=self.tags)
             if self.instance_count_unique_metric is not None:
-                self.check.gauge(self.instance_count_unique_metric, len(unique_instance_counts), tags=self.tags)
+                self.check.gauge(self.instance_count_unique_metric, unique_instance_count, tags=self.tags)
 
-        for counter in self.counters:
-            if counter.name not in counters:
-                self.logger.error(
-                    'Did not find expected counter `%s` of performance object `%s`', counter.name, self.name
-                )
-                counter.clear()
-            else:
-                counter.refresh(unique_instance_counts)
+    def refresh(self):
+        # Counters configuration should run only once in the current implementation.
+        #
+        # Removed non-default edge case support of dynamic addition/removal of configured performance
+        # counters. Even if a provider of a performance counter object, specified in our configuration,
+        # is not running, as long as its provider has been installed/deployed on the box, that is all
+        # needed for this integration. When the counter provider starts, Agent will discover its new
+        # instances and report them as metrics correspondingly.
+        #
+        # However, if the provider is installed AFTER the agent starts running, this implementation
+        # will not be able to automatically discover and use it. In some way it is a step back because
+        # this ability had been added in 7.34 among few other very important changes including detection
+        # of appearance and disappearance of performance counter instances, support for non-unique
+        # instances and  localization support. In this release detection of newly registered performance
+        # counters providers (and their objects and counters) is pulled out for the following reasons:
+        #    * It does not appear that the feature was needed by customers
+        #    * It adds non-insignificant complexity to the code
+        #    * It relies on very slow and resource hungry PdhEnumObjects(refresh=TRUE) function.
+        #    * It pretty much requires the agent process to be run as administrator or local system
+        #      user (otherwise will generate a handful of error messages in the event log every time
+        #      refresh run). Microsoft support confirmed it even though it is not documented.  It is
+        #      difficult to explain to customers.
+        #    * The logic of dynamically adding and removing counters handles to facilitate support appears
+        #      to may cause memory leaks in some counters (documented at least for ASP.NET e.g.).
+        # Perhaps in future if customers would really want that and are ready to deal with the caveats above
+        # we can add it back.
+
+        if not self.counters:
+            self._configure_counters()
+            if self.counters:
+                for counter in self.counters:
+                    counter.refresh()
 
     def clear(self):
         for counter in self.counters:
             counter.clear()
 
-    def _configure_counters(self, available_counters, available_instances):
-        if not available_counters:
-            return
-
+    def _configure_counters(self):
         if self.use_localized_counters:
             # https://docs.microsoft.com/en-us/windows/win32/api/pdh/nf-pdh-pdhaddcountera
             # https://mhammond.github.io/pywin32/win32pdh__AddCounter_meth.html
@@ -156,24 +229,21 @@ class PerfObject:
             # https://mhammond.github.io/pywin32/win32pdh__AddEnglishCounter_meth.html
             counter_selector = win32pdh.AddEnglishCounter
 
-        if available_instances:
-            counter_type = MultiCounter
+        # If a performance object and its counters are installed we will be able to determine its type.
+        # If they are not installed yet (see comment in refresh() method) counter type determination
+        # will fail. Moreover, it will continue to fail no matter how many times it is called even after
+        # the missing performance object and its counters have been installed.
+        counter_type = self._get_counters_type(counter_selector)
+        if counter_type == MultiCounter:
             self.has_multiple_instances = True
+        elif counter_type == SingleCounter:
+            self.has_multiple_instances = False
         else:
-            possible_path = construct_counter_path(
-                machine_name=self.connection.server, object_name=self.name, counter_name=available_counters[0]
+            raise Exception(
+                f'None of the specified counters for performance object `{self.name}` are installed.'
+                'If the object name is localized make sure `use_localized_counters` is true'
             )
 
-            # https://docs.microsoft.com/en-us/windows/win32/api/pdh/nf-pdh-pdhvalidatepatha
-            # https://mhammond.github.io/pywin32/win32pdh__ValidatePath_meth.html
-            if win32pdh.ValidatePath(possible_path) == 0:
-                counter_type = SingleCounter
-                self.has_multiple_instances = False
-            # Multi-instance counter with no instances presently
-            else:
-                counter_type = MultiCounter
-                self.has_multiple_instances = True
-
         tag_name = self.tag_name
         if self.has_multiple_instances:
             if not tag_name:
@@ -236,13 +306,49 @@ class PerfObject:
                         self.tags,
                         custom_transformers.get(counter_name),
                         tag_name,
+                        self.include_wildcards,
+                        self.duplicate_instances_exist,
+                        self,
                     )
 
         self.counters.extend(counters.values())
 
+    def _get_counters_type(self, counter_selector):
+        # Enumerate all counter to find if it is single or multiple instance counter
+        # The very virst iteration should be sufficienmt, just in case enumerate all
+        for i, entry in enumerate(self.counters_config, 1):
+            if not isinstance(entry, dict):
+                raise ConfigTypeError(
+                    f'Entry #{i} of option `counters` for performance object `{self.name}` must be a mapping'
+                )
+
+            for counter_name, _ in entry.items():
+                # Check for multi-instance counter path
+                possible_path = construct_counter_path(
+                    machine_name=self.connection.server,
+                    object_name=self.name,
+                    counter_name=counter_name,
+                    instance_name='*',
+                )
+                if validate_path(self.connection.query_handle, counter_selector, possible_path):
+                    return MultiCounter
+
+                # Check for single-instance counter path
+                possible_path = construct_counter_path(
+                    machine_name=self.connection.server,
+                    object_name=self.name,
+                    counter_name=counter_name,
+                )
+                if validate_path(self.connection.query_handle, counter_selector, possible_path):
+                    return SingleCounter
+
+        return None
+
     def get_custom_transformers(self):
         return {}
 
+    # At least one derived class (IIS CompatibilityPerfObject(PerfObject)) relies on this
+    # function call as callback for its own state management
     def _instance_excluded(self, instance):
         return self.exclude_pattern.search(instance) or (
             self.include_pattern is not None and not self.include_pattern.search(instance)
@@ -319,8 +425,8 @@ class CounterBase:
         # Counter requires at least 2 data points to return a meaningful value, see:
         # https://docs.microsoft.com/en-us/windows/win32/api/pdh/nf-pdh-pdhgetformattedcountervalue#remarks
         #
-        # https://mhammond.github.io/pywin32/error.html
-        if error.strerror != 'The data is not valid.':
+        # https://github.com/mhammond/pywin32/blob/main/win32/src/PyWinTypesmodule.cpp#L278
+        if error.winerror != PDH_NO_DATA:
             raise
 
         if instance is None:
@@ -359,7 +465,7 @@ class SingleCounter(CounterBase):
         else:
             self.transformer(value, tags=self.tags)
 
-    def refresh(self, _):
+    def refresh(self):
         if self.counter_handle is None:
             self.counter_handle = self.counter_selector(self.connection.query_handle, self.path)
 
@@ -395,11 +501,17 @@ class MultiCounter(CounterBase):
         tags,
         custom_transformer,
         tag_name,
+        include_wildcards,
+        duplicate_instances_exist,
+        perf_obj,
     ):
         super().__init__(
             name, config, check, connection, object_name, metric_prefix, counter_selector, tags, custom_transformer
         )
 
+        # Weak reference is used to avoid circular dependency and potentail memory leak
+        self.perf_obj_weakref = weakref.ref(perf_obj)
+
         self.tag_name = tag_name
 
         self.average = config.get('average', False)
@@ -426,127 +538,125 @@ class MultiCounter(CounterBase):
                 metric_name = f'{self.metric_name}.avg' if self.average else f'{self.metric_name}.sum'
                 self.aggregate_transformer = NATIVE_TRANSFORMERS[self.metric_type](check, metric_name, config)
 
+        self.include_wildcards = include_wildcards
+        self.duplicate_instances_exist = duplicate_instances_exist
+
+        self.total_instance_count = 0
+        self.monitored_instance_count = 0
+        self.unique_instance_count = 0
+
         # All monitored counter handles keyed by the instance name
-        self.instances = {}
+        self.instances = []
 
     def collect(self):
-        handles_with_data = 0
+        # Instance tracking is conducted for each counter although collected only from one.
+        # This counting is cheap but making it conditional here will only complicate this method
+        self.total_instance_count = 0
+        self.monitored_instance_count = 0
+        self.unique_instance_count = 0
+
         total = 0
-        for instance, counter_handles in self.instances.items():
-            # Some instances may not have a value yet, so we only
-            # use the ones that do for computing averages
-            instance_handles_with_data = len(counter_handles)
-            instance_total = 0
+        for counter_handle in self.instances:
+            try:
+                instance_items = get_counter_values(counter_handle, self.duplicate_instances_exist)
 
-            for i, counter_handle in enumerate(counter_handles):
-                try:
-                    value = get_counter_value(counter_handle)
-                except pywintypes.error as error:
-                    instance_handles_with_data -= 1
-                    self.handle_counter_value_error(error, format_instance(instance, i))
-                else:
-                    instance_total += value
+                for instance, value in instance_items.items():
+                    # Some instances may not have a value yet, so we only
+                    # use the ones that do for computing averages
+                    instance_non_unique_count = 0
+                    instance_total = 0
 
-            if not instance_handles_with_data:
-                continue
+                    self.total_instance_count += 1
+                    if self._instance_excluded(instance):
+                        continue
 
-            handles_with_data += instance_handles_with_data
-            total += instance_total
+                    self.unique_instance_count += 1
 
-            if self.aggregate != 'only':
-                tags = [f'{self.tag_name}:{instance}']
-                tags.extend(self.tags)
+                    # Enumerate non-unique instances
+                    if isinstance(value, list):
+                        for sub_value in value:
+                            instance_total += sub_value
+                            instance_non_unique_count += 1
+                    else:
+                        instance_total = value
+                        instance_non_unique_count += 1
 
-                if self.average:
-                    self.transformer(instance_total / instance_handles_with_data, tags=tags)
-                else:
-                    self.transformer(instance_total, tags=tags)
+                    if self.aggregate != 'only':
+                        tags = [f'{self.tag_name}:{instance}']
+                        tags.extend(self.tags)
 
-        if not handles_with_data:
+                        if self.average:
+                            self.transformer(instance_total / instance_non_unique_count, tags=tags)
+                        else:
+                            self.transformer(instance_total, tags=tags)
+
+                    if not instance_non_unique_count:
+                        continue
+
+                    self.monitored_instance_count += instance_non_unique_count
+                    total += instance_total
+
+            except pywintypes.error as error:
+                self.handle_counter_value_error(error)
+            except KeyError:
+                # To support IIS mocking tests for non-existing wildcard counters
+                pass
+
+        if not self.monitored_instance_count:
             return
 
         if self.aggregate is not False:
             if self.average:
-                self.aggregate_transformer(total / handles_with_data, tags=self.tags)
+                self.aggregate_transformer(total / self.monitored_instance_count, tags=self.tags)
             else:
                 self.aggregate_transformer(total, tags=self.tags)
 
-    def refresh(self, instance_counts):
-        old_instances = self.instances
-        new_instances = {}
+    def refresh(self):
+        # No need to create a counter handle or handles if they already exist. And since new
+        # counters for a performance object can be detected (without PdhEnumObjects(refresh=True))
+        # there is no point to make sure that all counters are available.
+        if self.instances and len(self.instances) > 0:
+            return
 
-        for instance, current_count in instance_counts.items():
-            if instance in old_instances:
-                counter_handles = old_instances.pop(instance)
-                new_instances[instance] = counter_handles
-                old_count = len(counter_handles)
+        self.instances = []
+        if self.include_wildcards is None:
+            self.include_wildcards = ["*"]
 
-                if current_count > old_count:
-                    for index in range(old_count, current_count):
-                        path = construct_counter_path(
-                            machine_name=self.connection.server,
-                            object_name=self.object_name,
-                            counter_name=self.name,
-                            instance_name=instance,
-                            instance_index=index,
-                        )
-                        counter_handle = self.counter_selector(self.connection.query_handle, path)
-                        counter_handles.append(counter_handle)
-                elif current_count < old_count:
-                    for _ in range(old_count - current_count):
-                        counter_handle = counter_handles.pop()
+        for _, pattern in enumerate(self.include_wildcards, 1):
+            path = construct_counter_path(
+                machine_name=self.connection.server,
+                object_name=self.object_name,
+                counter_name=self.name,
+                instance_name=pattern,
+            )
 
-                        try:
-                            # https://docs.microsoft.com/en-us/windows/win32/api/pdh/nf-pdh-pdhremovecounter
-                            # https://mhammond.github.io/pywin32/win32pdh__RemoveCounter_meth.html
-                            win32pdh.RemoveCounter(counter_handle)
-                        except Exception as e:
-                            self.logger.warning(
-                                'Unable to remove handle for instance `%s` of counter `%s` '
-                                'of performance object `%s`: %s',
-                                format_instance(instance, len(counter_handles)),
-                                self.name,
-                                self.object_name,
-                                e,
-                            )
-            else:
-                counter_handles = []
-                new_instances[instance] = counter_handles
-
-                for index in range(current_count):
-                    path = construct_counter_path(
-                        machine_name=self.connection.server,
-                        object_name=self.object_name,
-                        counter_name=self.name,
-                        instance_name=instance,
-                        instance_index=index,
-                    )
-                    counter_handle = self.counter_selector(self.connection.query_handle, path)
-                    counter_handles.append(counter_handle)
-
-        # Remove expired instances
-        self.clear()
-
-        self.instances = new_instances
+            counter_handle = self.counter_selector(self.connection.query_handle, path)
+            self.instances.append(counter_handle)
 
     def clear(self):
         if not self.instances:
             return
 
-        for instance, counter_handles in self.instances.items():
-            while counter_handles:
-                counter_handle = counter_handles.pop()
-                try:
-                    # https://docs.microsoft.com/en-us/windows/win32/api/pdh/nf-pdh-pdhremovecounter
-                    # https://mhammond.github.io/pywin32/win32pdh__RemoveCounter_meth.html
-                    win32pdh.RemoveCounter(counter_handle)
-                except Exception as e:
-                    self.logger.warning(
-                        'Unable to remove handle for instance `%s` of counter `%s` of performance object `%s`: %s',
-                        format_instance(instance, len(counter_handles)),
-                        self.name,
-                        self.object_name,
-                        e,
-                    )
+        for counter_handle in self.instances:
+            try:
+                # https://docs.microsoft.com/en-us/windows/win32/api/pdh/nf-pdh-pdhremovecounter
+                # https://mhammond.github.io/pywin32/win32pdh__RemoveCounter_meth.html
+                win32pdh.RemoveCounter(counter_handle)
+            except Exception as e:
+                self.logger.warning(
+                    'Unable to remove handle for counter `%s` of performance object `%s`: %s',
+                    self.name,
+                    self.object_name,
+                    e,
+                )
 
         self.instances.clear()
+
+    def get_instance_count(self):
+        return self.total_instance_count, self.monitored_instance_count, self.unique_instance_count
+
+    def _instance_excluded(self, instance):
+        # There is no need for additional validation since call to this object
+        # is always done from the live parent object (perf_obj)
+        perf_obj = self.perf_obj_weakref()
+        return perf_obj._instance_excluded(instance)
